//
// C++ <-> Lua expoter
// Magnus sterlind (magnus@dooz.se)
//
// Note, the code here also implements exporting member functions, which was only briefly mentioned in the article.
//

#pragma once

extern "C"
{
  #include <lua.h>
  #include <lauxlib.h>
}

#include <stdio.h>

#include <string>
#include <vector>
#include <exception>

#include <boost/any.hpp>
#include <boost/function_types/function_arity.hpp>
#include <boost/function_types/parameter_types.hpp>
#include <boost/function_types/result_type.hpp>
#include <boost/mpl/at.hpp>
#include <boost/mpl/for_each.hpp>
#include <boost/mpl/if.hpp> 
#include <boost/mpl/int.hpp>
#include <boost/mpl/size.hpp>
#include <boost/mpl/transform.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/type_traits/is_void.hpp> 
#include <boost/type_traits/remove_reference.hpp>

class LuaExporter
{
  typedef std::vector<boost::any> AnyParams;

  // Helper templates for making unique types from other types or from ints
  template<int T> struct Int2Type {}; 

  // Transform that removes the references from types in the sequence
  struct remove_ref {
    template <class T>
    struct apply {
      typedef typename boost::remove_reference<T>::type type;
    };
  };

  /**
  * Functions to push return values on the Lua stack
  */
  static void AddReturnValue(lua_State* lua_state, const bool value) {
    lua_pushboolean(lua_state, value ? 1 : 0);
  }

  static void AddReturnValue(lua_State* lua_state, const int value) {
    lua_pushinteger(lua_state, value);
  }

  static void AddReturnValue(lua_State* lua_state, const double value) {
    lua_pushnumber(lua_state, value);
  }

  static void AddReturnValue(lua_State* lua_state, const std::string& value) { 
    lua_pushlstring(lua_state, value.c_str(), value.length());
  }

  /**
  * The extractor is a functor that's called once for each type in the parameter type list. Each
  * application of operator() will call the correct overloaded Extract method based on the type 
  * of the current parameter, and try to extract the next value from the Lua stack
  * Parameters are stored in a vector<boost::any>
  */ 
  class Extractor
  {
  public:
    Extractor(lua_State* lua_state, AnyParams& params) : lua_state_(lua_state), params_(params), stack_index_(1) {}

    void Extract(const bool&) {
      if (!lua_isboolean(lua_state_, stack_index_)) {
        throw std::exception("Expected parameter of type boolean");
      }
      params_.push_back(lua_toboolean(lua_state_, stack_index_) == 1);
    }

    void Extract(const int&) {
      if (!lua_isnumber(lua_state_, stack_index_)) {
        throw std::exception("Expected parameter of type interger");
      }
      params_.push_back(lua_tointeger(lua_state_, stack_index_));
    }

    void Extract(const double&) {
      if (!lua_isnumber(lua_state_, stack_index_)) {
        throw std::exception("Expected parameter of type double");
      }
      // Note, you might want to change this to a cast to float instead of passing around doubles..
      params_.push_back(static_cast<double>(lua_tonumber(lua_state_, stack_index_)));
    }

    void Extract(const std::string&) { 
      if (!lua_isstring(lua_state_, stack_index_)) {
        throw std::exception("Expected parameter of type string");
      }
      params_.push_back(std::string(lua_tostring(lua_state_, stack_index_)));
    }

    template< typename T >
    void operator()(T t) {
      Extract(t);
      stack_index_++;
    }
  private:
    lua_State* lua_state_;
    AnyParams& params_;
    int stack_index_;
  };

  /**
  *	The executor is a class with operator() specialized on the class Int2Type<N>, where N is the
  * arity of the function we're calling. By specializing on the arity, we know the number of parameters
  * the function expects, and can extract these from the vector<boost::any> and cast to the correct type
  */ 
  template<class Arity, class ParamTypes, typename ResultType>
  struct Executor;

  // 0 parameters
  template<class ParamTypes, typename ResultType>
  struct Executor< Int2Type<0>, ParamTypes, ResultType> {
    template<class Fn> ResultType operator()(Fn& fn, const AnyParams& /*params*/) {
      return fn();
    }
  };

  // 1 parameter
  template<class ParamTypes, typename ResultType>
  struct Executor<Int2Type<1>, ParamTypes, ResultType> {
    template<class Fn> ResultType operator()(Fn& fn, const AnyParams& params) {
      return fn( 
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<0> >::type>(params[0]) );
    }
  };

  // 2 parameters
  template<class ParamTypes, typename ResultType>
  struct Executor<Int2Type<2>, ParamTypes, ResultType> {
    template<class Fn> ResultType operator()(Fn& fn, const AnyParams& params) {
      return fn( 
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<0> >::type>(params[0]),
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<1> >::type>(params[1]) );
    }
  };

  // 3 parameters
  template<class ParamTypes, typename ResultType>
  struct Executor< Int2Type<3>, ParamTypes, ResultType > {
    template<class Fn> ResultType operator()(Fn& fn, const AnyParams& params) {
      return fn( 
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<0> >::type>(params[0]),
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<1> >::type>(params[1]),
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<2> >::type>(params[2]) );
    }
  };

  template<class Arity, class ParamTypes, typename ResultType>
  struct ClassExecutor;

  // 0 parameters
  template<class ParamTypes, typename ResultType>
  struct ClassExecutor< Int2Type<0>, ParamTypes, ResultType> {
    template<class Obj, class MemFn> ResultType operator()(Obj& obj, MemFn& mem_fn, const AnyParams& /*params*/) {
      return (obj->*mem_fn)();
    }
  };

  // 1 parameter
  template<class ParamTypes, typename ResultType>
  struct ClassExecutor<Int2Type<1>, ParamTypes, ResultType> {
    template<class Obj, class MemFn> ResultType operator()(Obj& obj, MemFn& mem_fn, const AnyParams& params) {
      return (obj->*mem_fn)( 
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<0> >::type>(params[0]) );
    }
  };

  // 2 parameters
  template<class ParamTypes, typename ResultType>
  struct ClassExecutor<Int2Type<2>, ParamTypes, ResultType> {
    template<class Obj, class MemFn> ResultType operator()(Obj& obj, MemFn& mem_fn, const AnyParams& params) {
      return (obj->*mem_fn)( 
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<0> >::type>(params[0]),
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<1> >::type>(params[1]) );
    }
  };

  // 3 parameters
  template<class ParamTypes, typename ResultType>
  struct ClassExecutor< Int2Type<3>, ParamTypes, ResultType > {
    template<class Obj, class MemFn> ResultType operator()(Obj& obj, MemFn& mem_fn, const AnyParams& params) {
      return (obj->*mem_fn)( 
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<0> >::type>(params[0]),
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<1> >::type>(params[1]),
        boost::any_cast< boost::mpl::at<ParamTypes, boost::mpl::int_<2> >::type>(params[2]) );
    }
  };

  class Command {
  public:
    Command(const std::string& name) : name_(name) {}
    virtual ~Command() {}
  private:
    std::string name_;
  };

  template< class Fn >
  struct CommandT : public Command {
    CommandT(const std::string& name, const Fn& fn) 
      : Command(name)
      , fn_(fn) 
    {}

    // Remove the references from the parameter types
    typedef typename boost::mpl::transform< boost::function_types::parameter_types<Fn>, remove_ref>::type ParamTypes;

    // Store result type, and if the result type is a void or not
    typedef typename boost::function_types::result_type<Fn>::type ResultType;
    enum { IsVoidResult = boost::is_void<ResultType>::value };

    // Typedef for the executor (based on the arity of the function)
    typedef Executor< Int2Type<boost::function_types::function_arity<Fn>::value>, ParamTypes, ResultType > Executor;

    // Executer for void return type
    int Execute(lua_State* lua_state, Int2Type<true>) const {
      AnyParams params;
      try {
        boost::mpl::for_each<ParamTypes>(Extractor(lua_state, params));
      } catch( std::exception& /*ex*/) {
        //LOG_ERROR_LN("Error in extract");
        throw;
      }
      Executor f;
      f(fn_, params);
      return 0;
    }

    // Executer for non void return type
    int Execute(lua_State* lua_state, Int2Type<false>) const {
      AnyParams params;
      try {
        boost::mpl::for_each<ParamTypes>(Extractor(lua_state, params));
      } catch( std::exception& /*ex*/) {
        //LOG_ERROR_LN("Error in extract");
        throw;
      }
      Executor f;
      AddReturnValue(lua_state, f(fn_, params));
      return 1;
    }

    /**
    *	Called from lua. We've stored our this ptr in the c-closure associated with the function
    */ 
    static int callback(lua_State* lua_state){
      const CommandT* this_ptr = reinterpret_cast<CommandT*>(lua_tointeger(lua_state, lua_upvalueindex(1)));
      try {
        return this_ptr->Execute(lua_state, Int2Type<IsVoidResult>());
      } catch( std::exception& /*ex*/) {
        return 0;
      }
    }

    Fn fn_;
  };

  template< class T, class MemFn >
  struct CommandC : public Command {
    CommandC(const std::string& name, const T& obj, const MemFn& fn) 
      : Command(name)
      , obj_(obj)
      , mem_fn_(fn) 
    {}

    ~CommandC() {
    }

    // Remove the references from the parameter types
    typedef typename boost::mpl::transform< boost::function_types::parameter_types<MemFn>, remove_ref>::type ParamTypes1;

    // Remove the class from the parameter types
    typedef typename boost::mpl::pop_front<ParamTypes1>::type ParamTypes;

    // Store result type, and if the result type is a void or not
    typedef typename boost::function_types::result_type<MemFn>::type ResultType;
    enum { IsVoidResult = boost::is_void<ResultType>::value };

    // Typedef for the executor (based on the arity of the function)
    typedef ClassExecutor< Int2Type<boost::mpl::size<ParamTypes>::value>, ParamTypes, ResultType > Executor;

    // Executer for void return type
    int Execute(lua_State* lua_state, Int2Type<true>) {
      AnyParams params;
      try {
        boost::mpl::for_each<ParamTypes>(Extractor(lua_state, params));
      } catch (std::exception&) {
        throw;
      }
      Executor f;
      f(obj_, mem_fn_, params);
      return 0;
    }

    // Executer for non void return type
    int Execute(lua_State* lua_state, Int2Type<false>) {
      AnyParams params;
      try {
        boost::mpl::for_each<ParamTypes>(Extractor(lua_state, params));
      } catch (std::exception&) {
        throw;
      }
      Executor f;
      AddReturnValue(lua_state, f(obj_, mem_fn_, params));
      return 1;
    }

    static int callback(lua_State* lua_state){
      CommandC* this_ptr = reinterpret_cast<CommandC*>(lua_tointeger(lua_state, lua_upvalueindex(1)));
      try {
        return this_ptr->Execute(lua_state, Int2Type<IsVoidResult>());
      } catch (std::exception&) {
        return 0;
      }
    }

    T obj_;
    MemFn mem_fn_;
  };

public:
  LuaExporter() {
    lua_state_ = lua_open();
    if ( 0 == lua_state_) {
      throw std::exception("Unable to open lua state");
    }
  }

  ~LuaExporter() {
    lua_close(lua_state_);
  }

  /**
   * Register for a free function
   */ 
  template< class Fn >
  void Register(std::string function_name, const Fn& fn) {
    CommandT<Fn>* ptr = new CommandT<Fn>(function_name, fn);
    // Store the this pointer in the c-closure
    lua_pushinteger(lua_state_, reinterpret_cast<int>(ptr));
    lua_pushcclosure(lua_state_, &CommandT<Fn>::callback, 1);
    lua_setglobal(lua_state_, function_name.c_str());
    commands_.push_back(boost::shared_ptr<Command>(ptr));
  }

  /**
   * Register for a member function
   */ 
  template< class T, class MemFn >
  void Register(std::string function_name, MemFn mem_fn, T obj) {
    typedef CommandC<T, MemFn> Cmd;
    Cmd* ptr = new Cmd(function_name, obj, mem_fn);
    // Store the this pointer in the c-closure
    lua_pushinteger(lua_state_, reinterpret_cast<int>(ptr));
    lua_pushcclosure(lua_state_, &Cmd::callback, 1);
    lua_setglobal(lua_state_, function_name.c_str());
    commands_.push_back(boost::shared_ptr<Command>(ptr));
  }

  // TODO: Add unregister

  bool DoFile(const char* filename) {
    if( 0 != luaL_dofile(lua_state_, filename) ) {
      //LOG_ERROR_LN("Error running lua script: %s", filename);
      return false;
    }
    return true;
  }

private:
  std::vector<boost::shared_ptr<Command> > commands_;
  lua_State* lua_state_;
};
